package it.uniroma1.dis.jasmine.ui.utils;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.Observable;

import javax.swing.text.JTextComponent;

/**
 * Esegue il processo passato come parametro, scrivendo su un componente di testo il suo output e su un altro gli eventuali errori.
 * La classe estende Observable per notificare la fine dell'esecuzione.
 * 
 * @see Observable
 * @see Runnable
 */
public class ProcessRunner extends Observable implements Runnable {
	private JTextComponent out;
	private JTextComponent err;
	private String nusmvPath;
	private String smvFile;
	private Process process;
	private ProcessReader readOut;
	private ProcessReader readErr;
	/**
	 * @param out componente su cui scrivere l'output
	 * @param err componente su cui scriere gli errori
	 * @param nusmvPath path dell'eseguibile NuSMV
	 * @param smvFile path del file SMV
	 */
	public ProcessRunner(JTextComponent out, JTextComponent err, String nusmvPath, String smvFile) {
		super();
		this.out = out;
		this.err = err;
		this.nusmvPath = nusmvPath;
		this.smvFile = smvFile;
	}
	
	/**
	 * Richiede al thread di fermarsi il prima possibile.
	 */
	public void requestStop() {
		if (readOut != null)
			readOut.requestStop();
		if (readErr != null)
			readErr.requestStop();
		if (process != null)
			process.destroy();
	}
	
	/**
	 * Invia al processo eseguito una stringa.
	 * @param data
	 * @return
	 */
	public boolean write(String data) {
		if (process != null) {
			try {
				new PrintWriter(process.getOutputStream()).write(data);
				return true;
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		return false;
	}
	
	public String getProcessOutput() {
		return out.getText();
	}
	
	/**
	 * Crea il processo, si mette in ascolto sul suo canale di output e scrive sul componente di testo.
	 * Quando il processo termina o viene richiesto lo stop {@link #requestStop()}, notifica eventuali osservatori.
	 * In caso d'errore, notifica eventuali osservatori passando come parametro un oggetto {@link Exception}.
	 * @see java.lang.Thread#run()
	 */
	public void run() {
		out.setText("");
		if (err != null)
			err.setText("");
		try {
			process = Runtime.getRuntime().exec(nusmvPath + " " + smvFile);
			System.out.println(nusmvPath + " " + smvFile);
			
			readOut = new ProcessReader(process.getInputStream(), out);
			readOut.start();
			readErr = new ProcessReader(process.getErrorStream(), err);
			readErr.start();
			
			int exit = process.waitFor();
			System.out.println("Process uscito con: " + exit);
			
			if (readOut.isAlive())
				readOut.join();
			if (readErr.isAlive())
				readErr.join();
			
			setChanged();
			notifyObservers();
		} catch (Exception e) {
			e.printStackTrace();
			
			setChanged();
			notifyObservers(e);
		}
	}
	
	/**
	 * Legge dallo stream passato come parametro, scrivendo nel componente di testo.
	 */
	static class ProcessReader extends Thread {
		private JTextComponent text;
		private InputStream is;
		private boolean stopRequested;

		/**
		 * @param is
		 * @param text
		 */
		public ProcessReader(InputStream is, JTextComponent text) {
			super();
			this.is = is;
			this.text = text;
			this.stopRequested = false;
		}
		
		/**
		 * Richiede al thread di fermarsi il prima possibile.
		 */
		public void requestStop() {
			this.stopRequested = true;
		}
		
		/* (non-Javadoc)
		 * @see java.lang.Thread#run()
		 */
		@Override
		public void run() {
			if ((is == null) || (text == null))
				return;
			
			try {
				BufferedReader br = new BufferedReader(new InputStreamReader(is));
				String line;
				do {
					line = br.readLine();
					
					if (line != null) {
						text.setText(text.getText() + line + "\n");
						text.setCaretPosition(text.getText().length());
					
					}
				} while (!stopRequested && (line != null));
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}
}
